/*
  mmlib.c
  Interface library for mergemod & mergemem
  
  This file is part of mergemem by Philipp Richter & Philipp Reisner
  Created 1999-01-15 by Ulrich Neumerkel

  mergemem is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  mergemem is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with mergemem; see the file COPYING.  If not, write to
  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  This is a rather inefficient reference implementation based on
  mergemod version 0.12.  To become more efficient, more flexible
  ioctl calls are needed.
*/

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <asm/page.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>

#include "mmlib.h"
#include "../mergemod/mergemod.h"

static int 
m_fd = -1;


/*** TEMPORAL EMULATION ***/
/*** TEMPORAL EMULATION ***/
/*** TEMPORAL EMULATION ***/
/*** TEMPORAL EMULATION ***/

#define PATH_MAX 300

static size_t
randomize_size_t(size_t n)
{
  const int randomization_required = 1==2;
  if (randomization_required)
    {
      if (n > 0)
	{ /* simulate kernel limit */
	  size_t r = ( random() % n )  + 1;
	  assert(r <= n && r > 0);
	  return r;
	}
    }
  return n;
}

/*
static int
process_alive(pid_t pid)
{
  struct stat buf;
  char path[PATH_MAX+1];
  sprintf( path, "/proc/%d", pid);
  if ( stat(path, &buf) == -1 )
    return 0;
  else
    return 1;
}
*/

int
IOCTL_MERGEPAIRS( struct ioctl_mergepairs * argsp )
{
  size_t mergedpages, i, nel;
  MEQUALPAIR * pagepairs;
  if ( argsp == NULL ) /* invalid data area */
    {
      return -1;
    }
  if ( argsp -> pagepairs == NULL
       || argsp -> pagepairs + argsp -> nel == NULL ) /* invalid data area */
    {
      errno = EFAULT;
      return -1;
    }
  mergedpages = 0;
  nel = argsp -> nel;
  pagepairs = argsp -> pagepairs;

  nel = randomize_size_t(nel); /* arbitrary kernel limit */
  for (i = 0; i < nel; i++)
    {
      int retval;
      struct mergemem_mmem mm_mmem;
      mm_mmem.pid2 = pagepairs[i].spid;
      mm_mmem.addr2 = (unsigned long) pagepairs[i].sv_addr;
      mm_mmem.pid1 = pagepairs[i].dpid;
      mm_mmem.addr1 = (unsigned long) pagepairs[i].dv_addr;
      retval = ioctl(m_fd, MERGEMEM_MERGE_MEM, &mm_mmem);
      pagepairs[i].status = retval;
      if (retval == 0)
	mergedpages++;
    }
  argsp->wel = i;
  argsp->mergedpages = mergedpages;
  return 0;
}

int IOCTL_HASHPAGES( struct ioctl_hashpages *argsp )
{
  pid_t pid;
  MPAGEINFO *pageinfos;
  size_t nel, i;
  size_t hashedpages = 0;
  unsigned char * wbufferp, *endbufferp;
  KHASHMETHOD khashmethod;

  if ( argsp == NULL)
    return -1;

  pid = argsp -> pid;
  pageinfos = argsp -> pageinfos;
  nel = argsp -> nel;
  khashmethod = argsp -> khashmethod;

  switch ( argsp -> khashmethod )
    {
    case KHADDROTHALF:
    case KHCONST:
      break;
    case KHEXTERNAL:
      wbufferp = argsp -> pbufferp;

      if ( 1==2 && wbufferp != NULL ) /* pseudo code */
	{
	  /* if wbufferp points to a read only mapping, then free this
	     mapping and use a new mapping for the next pages */
	  wbufferp = NULL;
	}

      endbufferp = wbufferp + argsp -> size;
      if (endbufferp - wbufferp < PAGE_SIZE)
	{ /* not a single page can be written */
	  errno = EOVERFLOW; /* internal error */
	  /* if the kernel can itself provide/map the space its ok as well */
	  return -1;
	}
      argsp -> pagesize = PAGE_SIZE;
      break;
    default:
      errno = ENXIO;
      return -1; /* errno missing */
    }

  nel = randomize_size_t(nel); /* arbitrary kernel limit */
  for ( i = 0; i < nel; i++ )
    {
      /* insert here some better method to avoid same hash values for
         incompatible pages */

      const unsigned long virtualcachemask = 0;
      pageinfos[i].hash = ( (unsigned long) pageinfos[i].v_addr ) &
	virtualcachemask & PAGE_MASK; /* ignore insignificant address parts */

      /* get physical address */
      {
	struct mergemem_get_phys_addr mm_gpa;
	int retval;
	mm_gpa.pid = pid;
	mm_gpa.addr = (unsigned long) pageinfos[i].v_addr;
	retval =  ioctl(m_fd, MERGEMEM_GET_PHYS_ADDR, &mm_gpa);
#ifdef MERGEMEM_PERMISSION
	if (retval == MERGEMEM_PERMISSION)
	  {
	    argsp -> errno = EPERM;
	    return -1;
	  }
#endif
	if ( retval != 0)
	  {
	    pageinfos[i].p_addr = NULL;
	    pageinfos[i].count = 0;
	    continue;
	  }
	pageinfos[i].p_addr = (void *) mm_gpa.phys_addr;
      }

      /* get reference count and built in hash */
      {
	struct mergemem_chksum mm_chk;
	mm_chk.pid = pid;
	mm_chk.addr = (unsigned long) pageinfos[i].v_addr;
	if ( ioctl(m_fd, MERGEMEM_GEN_CHECKSUM, &mm_chk) == 0 )
	  {
	    if ( khashmethod == KHADDROTHALF)
		pageinfos[i].hash ^= mm_chk.chksum;
	    pageinfos[i].count = mm_chk.nrefs;
	  }
	else
	  {
	    pageinfos[i].p_addr = NULL;
	    pageinfos[i].count = 0;
	    continue;
	  }
      }
      if ( khashmethod == KHCONST )
	pageinfos[i].hash ^= m_hash_const(NULL,0);
      else if ( khashmethod == KHEXTERNAL )
	{
	  struct mergemem_get_page mgp;
	  int retval;

	  mgp.pid = pid;
	  mgp.addr = (unsigned long) pageinfos[i].v_addr;
	  mgp.page = wbufferp;
	  if (wbufferp == NULL)
	    { /* use mapping instead of copying */
	    }
	  retval = ioctl(m_fd, MERGEMEM_GET_PAGE, &mgp);
	  if ( retval )
	    {
	      pageinfos[i].p_addr = NULL;
	      pageinfos[i].count = 0;
	      continue;
	    }
	  wbufferp += PAGE_SIZE; /* Update only in this case! */
	  if ( wbufferp+PAGE_SIZE > endbufferp )
	    break;
	}
      hashedpages++;
    }

  argsp -> wel = i;
  argsp -> succwel = hashedpages;
  return 0;
}

/*** TEMPORAL EMULATION END ***/
/*** TEMPORAL EMULATION END ***/
/*** TEMPORAL EMULATION END ***/
/*** TEMPORAL EMULATION END ***/
/*** TEMPORAL EMULATION END ***/




/* test function */
static void
randomizearea(void *p, size_t len)
{
  const int randomization_required = 1 == 2;
  char * sp = (char *) p;
  size_t i;
  if (randomization_required)
    for( i = 0 ; i < len; i++ )
      sp[i] = random();
}

static unsigned long
(*m_hash_default)(const void* addr, size_t size) = NULL;


/* semi exported */
void
dosimpleioctl(void)
{
  int modver;
  ioctl(m_fd, MERGEMEM_CHECK_VER, &modver);
}


static ssize_t
do_initialization(void)
{
  int modver;
  if ( (m_fd = open("/dev/mergemem", O_RDONLY)) == -1 )
    {
      struct stat buf;
      if ( stat("/dev/mergemem", &buf) == -1 )
	errno = ENOPKG; /* package not installed */
      else
	errno = EPERM; /* exists, but cannot read */
      return -1;
    }
  if ( ioctl(m_fd, MERGEMEM_CHECK_VER, &modver) != 0 || modver != MOD_VERSION )
    {
      m_fd = -1;
      errno = EPROTO; /* version mismatch or similar */
      return -1;
    }
  m_hash_default = m_hash_addrothalf;
  return 0;
}

static ssize_t
ensure_initialization(void)
{
  if (m_fd != -1)
    return 0;
  return do_initialization();
}


#define PAGE_ALIGN(a) \
((void*)(((unsigned long)(a)+PAGE_SIZE-1) & ~(PAGE_SIZE-1)))

ssize_t
m_getpageinfos(  
	       const pid_t pid,          
	       const void * start,
	       const void * const end,
	       MPAGEINFO * const pageinfos,
	       const size_t nel,
	       const void * * const contp )
{
  size_t i = 0;
  struct ioctl_getpageinfos args;

  if ( ensure_initialization() < 0)
    return -1;

  start = PAGE_ALIGN(start);

  if ( start > end || nel == 0 )
    {
      errno = EINVAL;
      return -1;
    }
  args.pid = pid;
  args.start = start;
  args.end = end;

  while ( i < nel  )
    {
      args.pageinfos = pageinfos+i;
      args.nel = nel-i;
      if ( ioctl(m_fd, MERGEMEM_GET_PAGEINFOS, &args) )
	return -1;
      i += args.wel;
      if ( args.cont == NULL )
	break;
      args.start = args.cont;
    }
  if(contp) *contp = args.cont;  /* MOD */
  return i;
}



ssize_t
m_hashpages(
	    const pid_t pid,
	    MPAGEINFO * const pageinfos,
	    const size_t nel,
	    unsigned long (*hashfunction)(const void* addr, size_t size) )
{
  ssize_t hashedpages, i;
  struct ioctl_hashpages args;


#define MAXPAGES 11

  unsigned char pagearray [PAGE_SIZE*MAXPAGES];
  randomizearea(pagearray, sizeof(pagearray));

  if ( ensure_initialization() < 0)
    return -1;
  if ( nel == 0 )
    {
      errno = EINVAL;
      return -1;
    }
  if ( hashfunction == NULL )
    hashfunction = m_hash_default; /* current default */

  hashedpages = 0;
  
  args.pid = pid;
  if ( hashfunction == m_hash_addrothalf )
    args.khashmethod = KHADDROTHALF;
  else if ( hashfunction == m_hash_const )
    args.khashmethod = KHCONST;
  else
    {
      args.khashmethod = KHEXTERNAL;
      args.pbufferp = pagearray;
      args.size = sizeof(pagearray);
    }
  i = 0;
  while (i < nel)
    {
      args.pageinfos = pageinfos+i;
      if ( args.khashmethod == KHEXTERNAL
	   && nel - i >= MAXPAGES )
	args.nel = MAXPAGES;
      else
	args.nel = nel - i;
      if ( IOCTL_HASHPAGES( & args) )
	  return -1;
      if (args.khashmethod == KHEXTERNAL)
	{
	  size_t j;
	  const unsigned char * pagestart = args.pbufferp;
	  /* the kernel might have chosen a better location */
	  size_t pagesize = args.pagesize;
	  for (j = 0; j < args.wel; j++)
	    {
	      if ( pageinfos[j+i].count != 0)
		{
		  pageinfos[j+i].hash ^=
		    (*hashfunction)(pagestart,pagesize);
		  pagestart += pagesize;
		}
	    }
	}
      i += args.wel;
      hashedpages += args.succwel;
    }
  if ( args.khashmethod == KHEXTERNAL
       && args.pbufferp != pagearray )
    {
      /* the kernel changed pbufferp to make it point to its own
         area, now remove this map from our process */
      args.nel = 0;
      if ( IOCTL_HASHPAGES( & args) )
	return -1;
    }
  randomizearea(pagearray, sizeof(pagearray));      
  return hashedpages;
}


ssize_t
m_merge( MEQUALPAIR * const pagepairs, const size_t nel )
{
  ssize_t mergedpages, i;

  if ( ensure_initialization() < 0)
    return -1;
  mergedpages = 0;
  i = 0;
  while ( i < nel )
    {
      struct ioctl_mergepairs args;

      args.pagepairs = pagepairs+i;
      args.nel = nel - i;
      if (  IOCTL_MERGEPAIRS( &args ) )
	return -1;
      i += args.wel;
      mergedpages += args.mergedpages;
    }
  return mergedpages;
}


ssize_t
m_set_hash_default( unsigned long (* const hashfunction)(const void* addr, size_t size) )
{
  if ( ensure_initialization() < 0 )
    return -1;
  if ( hashfunction == NULL )
    {
      errno = EINVAL;
      return -1;
    }
  m_hash_default = hashfunction;
  return 0;
}

/* Predefined hashfunctions */

#define BITS_OF_UNSIGNEDLONG (sizeof(unsigned long)*8)

unsigned long
m_hash_addrothalf(const void* addr, const size_t size)
{
  unsigned long * page = (unsigned long *) addr;
  unsigned long * end = (unsigned long *) ( (char *) addr + size/2 );
  unsigned long chksum = 0;
  while (page < end)
    {
      chksum = (chksum << 3) + (chksum >> (BITS_OF_UNSIGNEDLONG-3)) + *page;
      page++;
    }
  return chksum;
}

unsigned long
m_libhash_addrothalf(const void* addr, size_t size)
{
  return m_hash_addrothalf(addr, size);
}

unsigned long
m_hash_const(const void* addr, size_t size)
{
  return 27;
}

unsigned long
m_libhash_const(const void* addr, size_t size)
{
  return m_hash_const(addr, size);
}
